@@ -2,7 +2,7 @@
 
   vm.h -
 
-  $Author: nobu $
+  $Author: k0kubun $
   created at: 04/01/01 16:56:59 JST
 
   Copyright (C) 2004-2007 Koichi Sasada
@@ -27,7 +27,7 @@ typedef rb_iseq_t *ISEQ;
 #if VMDEBUG > 0
 #define debugs printf
 #define DEBUG_ENTER_INSN(insn) \
-  rb_vmdebug_debug_print_pre(th, GET_CFP());
+    rb_vmdebug_debug_print_pre(ec, GET_CFP(), GET_PC());
 
 #if OPT_STACK_CACHING
 #define SC_REGS() , reg_a, reg_b
@@ -60,22 +60,29 @@ error !
 
 #define INSN_ENTRY(insn) \
   static rb_control_frame_t * \
-    FUNC_FASTCALL(LABEL(insn))(rb_thread_t *th, rb_control_frame_t *reg_cfp) {
+    FUNC_FASTCALL(LABEL(insn))(rb_execution_context_t *ec, rb_control_frame_t *reg_cfp) {
 
 #define END_INSN(insn) return reg_cfp;}
 
 #define NEXT_INSN() return reg_cfp;
 
+#define START_OF_ORIGINAL_INSN(x) /* ignore */
+#define DISPATCH_ORIGINAL_INSN(x) return LABEL(x)(ec, reg_cfp);
+
 /************************************************/
 #elif OPT_TOKEN_THREADED_CODE || OPT_DIRECT_THREADED_CODE
 /* threaded code with gcc */
 
 #define LABEL(x)  INSN_LABEL_##x
 #define ELABEL(x) INSN_ELABEL_##x
-#define LABEL_PTR(x) &&LABEL(x)
-
-#define INSN_ENTRY_SIG(insn)
+#define LABEL_PTR(x) RB_GNUC_EXTENSION(&&LABEL(x))
 
+#define INSN_ENTRY_SIG(insn) \
+  if (0) fprintf(stderr, "exec: %s@(%d, %d)@%s:%d\n", #insn, \
+		 (int)(reg_pc - reg_cfp->iseq->body->iseq_encoded), \
+		 (int)(reg_cfp->pc - reg_cfp->iseq->body->iseq_encoded), \
+		 RSTRING_PTR(rb_iseq_path(reg_cfp->iseq)), \
+		 (int)(rb_iseq_line_no(reg_cfp->iseq, reg_pc - reg_cfp->iseq->body->iseq_encoded)));
 
 #define INSN_DISPATCH_SIG(insn)
 
@@ -91,17 +98,15 @@ error !
 #else
 #define DISPATCH_ARCH_DEPEND_WAY(addr) \
 				/* do nothing */
-
 #endif
 
-
 /**********************************/
 #if OPT_DIRECT_THREADED_CODE
 
 /* for GCC 3.4.x */
 #define TC_DISPATCH(insn) \
   INSN_DISPATCH_SIG(insn); \
-  goto *(void const *)GET_CURRENT_INSN(); \
+  RB_GNUC_EXTENSION_BLOCK(goto *(void const *)GET_CURRENT_INSN()); \
   ;
 
 #else
@@ -110,7 +115,7 @@ error !
 #define TC_DISPATCH(insn)  \
   DISPATCH_ARCH_DEPEND_WAY(insns_address_table[GET_CURRENT_INSN()]); \
   INSN_DISPATCH_SIG(insn); \
-  goto *insns_address_table[GET_CURRENT_INSN()]; \
+  RB_GNUC_EXTENSION_BLOCK(goto *insns_address_table[GET_CURRENT_INSN()]); \
   rb_bug("tc error");
 
 
@@ -130,6 +135,9 @@ error !
 
 #define NEXT_INSN() TC_DISPATCH(__NEXT_INSN__)
 
+#define START_OF_ORIGINAL_INSN(x) start_of_##x:
+#define DISPATCH_ORIGINAL_INSN(x) goto  start_of_##x;
+
 /************************************************/
 #else /* no threaded code */
 /* most common method */
@@ -141,7 +149,6 @@ case BIN(insn):
   DEBUG_END_INSN();                           \
   break;
 
-
 #define INSN_DISPATCH()         \
   while (1) {			\
     switch (GET_CURRENT_INSN()) {
@@ -155,26 +162,53 @@ default:                        \
 
 #define NEXT_INSN() goto first
 
+#define START_OF_ORIGINAL_INSN(x) start_of_##x:
+#define DISPATCH_ORIGINAL_INSN(x) goto  start_of_##x;
+
 #endif
 
-#define VM_SP_CNT(th, sp) ((sp) - (th)->stack)
+#ifdef MJIT_HEADER
+#define EXEC_EC_CFP() do { \
+    VM_ENV_FLAGS_SET(ec->cfp->ep, VM_FRAME_FLAG_FINISH); \
+    val = vm_exec(ec); \
+} while (0)
+#else
+#define EXEC_EC_CFP() do { \
+    RESTORE_REGS(); \
+    NEXT_INSN(); \
+} while (0)
+#endif
 
+#define VM_SP_CNT(ec, sp) ((sp) - (ec)->vm_stack)
+
+#ifdef MJIT_HEADER
+#define THROW_EXCEPTION(exc) do { \
+    ec->errinfo = (VALUE)(exc); \
+    EC_JUMP_TAG(ec, ec->tag->state); \
+} while (0)
+#else
 #if OPT_CALL_THREADED_CODE
 #define THROW_EXCEPTION(exc) do { \
-    th->errinfo = (VALUE)(exc); \
+    ec->errinfo = (VALUE)(exc); \
     return 0; \
 } while (0)
 #else
 #define THROW_EXCEPTION(exc) return (VALUE)(exc)
 #endif
+#endif
 
 #define SCREG(r) (reg_##r)
-#define CHECK_VM_STACK_OVERFLOW_FOR_INSN(cfp, margin) do { \
-    if (((rb_control_frame_t *)(((cfp)->sp) + (margin)) + 1) >= (cfp)) { \
-	rb_bug("CHECK_VM_STACK_OVERFLOW_FOR_INSN: should not overflow here. " \
-	       "Please contact ruby-core/dev with your (a part of) script. " \
-	       "This check will be removed soon."); \
-    } \
-} while (0)
+
+#define VM_DEBUG_STACKOVERFLOW 0
+
+#if VM_DEBUG_STACKOVERFLOW
+#define CHECK_VM_STACK_OVERFLOW_FOR_INSN(cfp, margin) \
+    WHEN_VM_STACK_OVERFLOWED(cfp, (cfp)->sp, margin) vm_stack_overflow_for_insn()
+#else
+#define CHECK_VM_STACK_OVERFLOW_FOR_INSN(cfp, margin)
+#endif
+
+#define INSN_LABEL2(insn, name) INSN_LABEL_ ## insn ## _ ## name
+#define INSN_LABEL(x) INSN_LABEL2(NAME_OF_CURRENT_INSN, x)
 
 #endif /* RUBY_VM_EXEC_H */
